@@ -18,7 +18,9 @@ import aiomqtt
 from aiomqtt import MqttCodeError, MqttError, TLSParameters
 
 from roborock.callbacks import CallbackMap
+from roborock.diagnostics import Diagnostics
 
+from .health_manager import HealthManager
 from .session import MqttParams, MqttSession, MqttSessionException, MqttSessionUnauthorized
 
 _LOGGER = logging.getLogger(__name__)
@@ -75,12 +77,19 @@ class RoborockMqttSession(MqttSession):
         self._connection_task: asyncio.Task[None] | None = None
         self._topic_idle_timeout = topic_idle_timeout
         self._idle_timers: dict[str, asyncio.Task[None]] = {}
+        self._diagnostics = params.diagnostics
+        self._health_manager = HealthManager(self.restart)
 
     @property
     def connected(self) -> bool:
         """True if the session is connected to the broker."""
         return self._healthy
 
+    @property
+    def health_manager(self) -> HealthManager:
+        """Return the health manager for the session."""
+        return self._health_manager
+
     async def start(self) -> None:
         """Start the MQTT session.
 
@@ -89,24 +98,30 @@ class RoborockMqttSession(MqttSession):
         handle the failure and retry if desired itself. Once connected,
         the session will retry connecting in the background.
         """
+        self._diagnostics.increment("start_attempt")
         start_future: asyncio.Future[None] = asyncio.Future()
         loop = asyncio.get_event_loop()
         self._reconnect_task = loop.create_task(self._run_reconnect_loop(start_future))
         try:
             await start_future
         except MqttCodeError as err:
+            self._diagnostics.increment(f"start_failure:{err.rc}")
             if err.rc == MqttReasonCode.RC_ERROR_UNAUTHORIZED:
                 raise MqttSessionUnauthorized(f"Authorization error starting MQTT session: {err}") from err
             raise MqttSessionException(f"Error starting MQTT session: {err}") from err
         except MqttError as err:
+            self._diagnostics.increment("start_failure:unknown")
             raise MqttSessionException(f"Error starting MQTT session: {err}") from err
         except Exception as err:
+            self._diagnostics.increment("start_failure:uncaught")
             raise MqttSessionException(f"Unexpected error starting session: {err}") from err
         else:
+            self._diagnostics.increment("start_success")
             _LOGGER.debug("MQTT session started successfully")
 
     async def close(self) -> None:
         """Cancels the MQTT loop and shutdown the client library."""
+        self._diagnostics.increment("close")
         self._stop = True
         tasks = [task for task in [self._connection_task, self._reconnect_task, *self._idle_timers.values()] if task]
         self._connection_task = None
@@ -129,6 +144,7 @@ class RoborockMqttSession(MqttSession):
         the reconnect loop. This is a no-op if there is no active connection.
         """
         _LOGGER.info("Forcing MQTT session restart")
+        self._diagnostics.increment("restart")
         if self._connection_task:
             self._connection_task.cancel()
         else:
@@ -137,6 +153,7 @@ class RoborockMqttSession(MqttSession):
     async def _run_reconnect_loop(self, start_future: asyncio.Future[None] | None) -> None:
         """Run the MQTT loop."""
         _LOGGER.info("Starting MQTT session")
+        self._diagnostics.increment("start_loop")
         while True:
             try:
                 self._connection_task = asyncio.create_task(self._run_connection(start_future))
@@ -157,6 +174,7 @@ class RoborockMqttSession(MqttSession):
                 _LOGGER.debug("MQTT session closed, stopping retry loop")
                 return
             _LOGGER.info("MQTT session disconnected, retrying in %s seconds", self._backoff.total_seconds())
+            self._diagnostics.increment("reconnect_wait")
             await asyncio.sleep(self._backoff.total_seconds())
             self._backoff = min(self._backoff * BACKOFF_MULTIPLIER, MAX_BACKOFF_INTERVAL)
 
@@ -168,17 +186,19 @@ class RoborockMqttSession(MqttSession):
         is lost, this method will exit.
         """
         try:
-            async with self._mqtt_client(self._params) as client:
-                self._backoff = MIN_BACKOFF_INTERVAL
-                self._healthy = True
-                _LOGGER.info("MQTT Session connected.")
-                if start_future and not start_future.done():
-                    start_future.set_result(None)
+            with self._diagnostics.timer("connection"):
+                async with self._mqtt_client(self._params) as client:
+                    self._backoff = MIN_BACKOFF_INTERVAL
+                    self._healthy = True
+                    _LOGGER.info("MQTT Session connected.")
+                    if start_future and not start_future.done():
+                        start_future.set_result(None)
 
-                _LOGGER.debug("Processing MQTT messages")
-                async for message in client.messages:
-                    _LOGGER.debug("Received message: %s", message)
-                    self._listeners(message.topic.value, message.payload)
+                    _LOGGER.debug("Processing MQTT messages")
+                    async for message in client.messages:
+                        _LOGGER.debug("Received message: %s", message)
+                        with self._diagnostics.timer("dispatch_message"):
+                            self._listeners(message.topic.value, message.payload)
         except MqttError as err:
             if start_future and not start_future.done():
                 _LOGGER.info("MQTT error starting session: %s", err)
@@ -220,6 +240,7 @@ class RoborockMqttSession(MqttSession):
                 async with self._client_lock:
                     self._client = client
                     for topic in self._client_subscribed_topics:
+                        self._diagnostics.increment("resubscribe")
                         _LOGGER.debug("Re-establishing subscription to topic %s", topic)
                         # TODO: If this fails it will break the whole connection. Make
                         # this retry again in the background with backoff.
@@ -244,6 +265,7 @@ class RoborockMqttSession(MqttSession):
 
         # If there is an idle timer for this topic, cancel it (reuse subscription)
         if idle_timer := self._idle_timers.pop(topic, None):
+            self._diagnostics.increment("unsubscribe_idle_cancel")
             idle_timer.cancel()
             _LOGGER.debug("Cancelled idle timer for topic %s (reused subscription)", topic)
 
@@ -255,13 +277,15 @@ class RoborockMqttSession(MqttSession):
                 if self._client:
                     _LOGGER.debug("Establishing subscription to topic %s", topic)
                     try:
-                        await self._client.subscribe(topic)
+                        with self._diagnostics.timer("subscribe"):
+                            await self._client.subscribe(topic)
                     except MqttError as err:
                         # Clean up the callback if subscription fails
                         unsub()
                         self._client_subscribed_topics.discard(topic)
                         raise MqttSessionException(f"Error subscribing to topic: {err}") from err
                 else:
+                    self._diagnostics.increment("subscribe_pending")
                     _LOGGER.debug("Client not connected, will establish subscription later")
 
         def schedule_unsubscribe() -> None:
@@ -294,10 +318,11 @@ class RoborockMqttSession(MqttSession):
             self._idle_timers[topic] = task
 
         def delayed_unsub():
+            self._diagnostics.increment("unsubscribe")
             unsub()  # Remove the callback from CallbackMap
             # If no more callbacks for this topic, start idle timer
             if not self._listeners.get_callbacks(topic):
-                _LOGGER.debug("Unsubscribing topic %s, starting idle timer", topic)
+                self._diagnostics.increment("unsubscribe_idle_start")
                 schedule_unsubscribe()
             else:
                 _LOGGER.debug("Unsubscribing topic %s, still have active callbacks", topic)
@@ -313,7 +338,8 @@ class RoborockMqttSession(MqttSession):
                 raise MqttSessionException("Could not publish message, MQTT client not connected")
             client = self._client
         try:
-            await client.publish(topic, message)
+            with self._diagnostics.timer("publish"):
+                await client.publish(topic, message)
         except MqttError as err:
             raise MqttSessionException(f"Error publishing message: {err}") from err
 
@@ -326,21 +352,28 @@ class LazyMqttSession(MqttSession):
     is made.
     """
 
-    def __init__(self, session: RoborockMqttSession) -> None:
+    def __init__(self, session: RoborockMqttSession, diagnostics: Diagnostics) -> None:
         """Initialize the lazy session with an existing session."""
         self._lock = asyncio.Lock()
         self._started = False
         self._session = session
+        self._diagnostics = diagnostics
 
     @property
     def connected(self) -> bool:
         """True if the session is connected to the broker."""
         return self._session.connected
 
+    @property
+    def health_manager(self) -> HealthManager:
+        """Return the health manager for the session."""
+        return self._session.health_manager
+
     async def _maybe_start(self) -> None:
         """Start the MQTT session if not already started."""
         async with self._lock:
             if not self._started:
+                self._diagnostics.increment("start")
                 await self._session.start()
                 self._started = True
 
@@ -391,4 +424,4 @@ async def create_lazy_mqtt_session(params: MqttParams) -> MqttSession:
     This function is a factory for creating an MQTT session that will
     only connect when the first attempt to subscribe or publish is made.
     """
-    return LazyMqttSession(RoborockMqttSession(params))
+    return LazyMqttSession(RoborockMqttSession(params), diagnostics=params.diagnostics.subkey("lazy_mqtt"))
